#include "operator.hh"
#include "waves.h"

void Operator::Modulate(Operator &o)
{
	o.source    = this;
	o.chaintype = mod_freq;
	target = &o;
}

void Operator::AddTo(Operator &o)
{
	o.source    = this;
	o.chaintype = mod_add;
	target = &o;
}

void Operator::Unlink(void)
{
	target = 0;
	source = 0;
}

const unsigned Operator::multiplyfreq(unsigned freq) const
{
	switch(freqmul)
	{
		case 0: return freq/2;
		case 1: return freq;
		case 11: return freq * 10;
		case 13: return freq * 12;
		case 14: return freq * 15;
	}
	return freq * freqmul;
}

void Operator::Write20(unsigned char value)
{
	tremolo = value & 0x80;
	vibrato = value & 0x40;
	dosustain=value & 0x20;
	KSR =     value & 0x10;
	freqmul = value & 15;
}

void Operator::Write40(unsigned char value)
{
	KSL         = value >> 6;
	outputlevel = value & 63;
}

void Operator::Write60(unsigned char value)
{
	attack   = value >> 4;
	decay    = value & 15;
}
	
void Operator::Write80(unsigned char value)
{
	sustain  = value >> 4;
	release  = value & 15;
}

unsigned OperatorPortOffset(unsigned channel)
{
	unsigned ofs = 0x100 * (channel / 18);
	register unsigned chan18 = channel % 18;
	return ofs + (chan18%6) + 2 * (chan18/6);
}

unsigned PortOffsetOperator(unsigned offset)
{
	unsigned chan = 18 * (offset / 0x100);
	register unsigned low = offset & 0x1F;
	
	chan = low;
	
	if(low >= 0x08)chan -= 2;
	if(low >= 0x10)chan -= 2;
	
	if(chan >= 36)chan = 36-1;
	
	return chan;
}

static const double PII = 3.141592653589793;

double Operator::Produce(unsigned waveno, double freq,
                         double angle, double shift)
{
	double result = waves[waveno&7] (freq * angle + shift);
	
	if(target)
	{
		if(chaintype == mod_add)
			result += target->Produce(waveno, freq, angle, 0.0);
		else
			result = target->Produce(waveno, freq, angle, PII*result);
	}
	return result;
}
